@@ -7,6 +7,11 @@ class Error(Exception):
     Base exception type for stages
     """
 
+    def __init__(self, *args, **kwargs):
+        super().__init__(*args, **kwargs)
+        self._pipeline_stage = None
+        self._exception_cause = None
+
     def set_stage(self, stage: str):
         self._pipeline_stage = stage
 
@@ -18,7 +23,9 @@ class Error(Exception):
         Set the original exception (if any) that has generated this error,
         equivalent to `explicit exception chaining <https://www.python.org/dev/peps/pep-3134/#explicit-exception-chaining>`_
         """
+        # they are both necessary, in particular the second one is to preserve the cause from child thread exception
         self.__cause__ = exception
+        self._exception_cause = exception
         return self
 
     def get_exception(self) -> Optional[Exception]:
@@ -26,7 +33,7 @@ class Error(Exception):
         Get the original exception (if any) that has generated this error,
         equivalent to the `__cause__ <https://www.python.org/dev/peps/pep-3134/#explicit-exception-chaining>`_ attribute
         """
-        return self.__cause__ or None
+        return self.__cause__ or self._exception_cause
 
 
 class SoftError(Error):
@@ -43,3 +50,13 @@ class CriticalError(Error):
     """
 
     pass
+
+
+class RetryError(SoftError):
+    """
+    A stage processing an input item may raise some Exception for which the stage must try to reprocess the item using
+    an exponential backoff retry strategy. After several attempts a RetryError is raised and
+    that stage is skipped for that item
+    """
+
+    pass
