@@ -8,10 +8,20 @@ from concurrent.futures.thread import ThreadPoolExecutor
 from multiprocessing import Manager
 from queue import Queue
 from threading import Thread, Event
-from typing import Generator, Optional, Callable, Sequence, Mapping, Any
+from typing import (
+    Generator,
+    Optional,
+    Callable,
+    Sequence,
+    Mapping,
+    Any,
+    Union,
+    Tuple,
+    Type,
+)
 
 from smartpipeline.defaults import CONCURRENCY_WAIT, MAX_QUEUES_SIZE
-from smartpipeline.error.handling import ErrorManager
+from smartpipeline.error.handling import ErrorManager, RetryManager
 from smartpipeline.containers import (
     SourceContainer,
     StageContainer,
@@ -260,6 +270,12 @@ class Pipeline:
                 container.queues_join()
                 while not container.queues_empty():
                     time.sleep(wait_seconds)
+            if (
+                not isinstance(container, ConcurrentStageContainer)
+                or not container.parallel
+            ):
+                container.stage.on_end()
+        self._error_manager.on_end()
         _logger.debug("Termination done")
 
     def _all_terminated(self) -> bool:
@@ -384,7 +400,14 @@ class Pipeline:
         executor.submit(_waiter)
 
     def _build_container(
-        self, name: str, stage: StageType, concurrency: int, parallel: bool
+        self,
+        name: str,
+        stage: StageType,
+        concurrency: int,
+        parallel: bool,
+        retryable_errors: Tuple[Type[Exception], ...],
+        max_retries: int,
+        backoff: Union[float, int],
     ) -> BaseContainer:
         """
         Get a new container instance according to the pipeline configuration
@@ -393,14 +416,20 @@ class Pipeline:
         :param stage: A stage instance
         :param concurrency: Number of concurrent stage executions, if 0 then just create the non-concurrent containers
         :param parallel: If True use multiprocessing, otherwise threads
+        :param retryable_errors: list of exceptions for which the stage applies an exponential backoff strategy. When the maximum number of retries is hit, a :class:`.error.exceptions.RetryError` is raised
+        :param max_retries: maximum number of retries for the stage before raising a RetryError(SoftError) (default 0)
+        :param backoff: backoff factor for the `exponential backoff strategy` used by the stage when it raises one of the exceptions declared in `retryable_errors` param (default 0.0)
         """
+        retry_manager = RetryManager(
+            backoff=backoff, max_retries=max_retries, retryable_errors=retryable_errors
+        )
         if concurrency <= 0:
             constructor = (
                 BatchStageContainer if isinstance(stage, BatchStage) else StageContainer
             )
             # if not concurrent we must explicitly finalize initialization of this single stage object
             stage.on_start()
-            return constructor(name, stage, self._error_manager)
+            return constructor(name, stage, self._error_manager, retry_manager)
         else:
             constructor = (
                 BatchConcurrentStageContainer
@@ -412,6 +441,7 @@ class Pipeline:
                     name,
                     stage,
                     self._error_manager,
+                    retry_manager,
                     self._new_mp_queue,
                     self._new_mp_counter,
                     self._new_mp_event,
@@ -426,6 +456,7 @@ class Pipeline:
                     name,
                     stage,
                     self._error_manager,
+                    retry_manager,
                     self._new_queue,
                     self._new_counter,
                     self._new_event,
@@ -440,7 +471,14 @@ class Pipeline:
         return self._containers.get(name).stage
 
     def append_stage(
-        self, name: str, stage: StageType, concurrency: int = 0, parallel: bool = False,
+        self,
+        name: str,
+        stage: StageType,
+        concurrency: int = 0,
+        parallel: bool = False,
+        retryable_errors: Tuple[Type[Exception], ...] = tuple(),
+        max_retries: int = 0,
+        backoff: Union[float, int] = 0.0,
     ) -> Pipeline:
         """
         Append a stage to the pipeline just after the last one appended, or after the source if it is the first stage
@@ -449,6 +487,9 @@ class Pipeline:
         :param stage: Instance of a stage
         :param concurrency: Number of concurrent stage executions, if 0 then threads/processes won't be involved for this stage
         :param parallel: If True use multiprocessing, otherwise threads
+        :param retryable_errors: list of exceptions for which the stage applies an exponential backoff strategy. When the maximum number of retries is hit, a :class:`.error.exceptions.RetryError` is raised
+        :param max_retries: maximum number of retries for the stage before raising a RetryError(SoftError) (default 0)
+        :param backoff: backoff factor for the `exponential backoff strategy` used by the stage when it raises one of the exceptions declared in `retryable_errors` param (default 0.0)
         """
         self._executors_ready = False
         # FIXME here we force a BatchStage to run on a thread, but we would leave it on the main thread
@@ -456,7 +497,10 @@ class Pipeline:
             parallel = False
             concurrency = 1
         self._check_stage_name(name)
-        container = self._build_container(name, stage, concurrency, parallel)
+        self._check_retries_params(retryable_errors, max_retries, backoff)
+        container = self._build_container(
+            name, stage, concurrency, parallel, retryable_errors, max_retries, backoff
+        )
         if concurrency > 0:
             # if it is concurrent and it is the first stage, make the source working on a output queue
             if not self._containers:
@@ -475,6 +519,9 @@ class Pipeline:
         kwargs: Mapping = None,
         concurrency: int = 0,
         parallel: bool = False,
+        retryable_errors: Tuple[Type[Exception], ...] = tuple(),
+        max_retries: int = 0,
+        backoff: Union[float, int] = 0.0,
     ) -> Pipeline:
         """
         Append a stage class to the pipeline just after the last one appended, or after the source if it is the first stage.
@@ -486,6 +533,9 @@ class Pipeline:
         :param kwargs: Dictionary of keyed arguments for the stage constructor
         :param concurrency: Number of concurrent stage executions, if 0 then threads/processes won't be involved for this stage
         :param parallel: If True use multiprocessing, otherwise threads
+        :param retryable_errors: list of exceptions for which the stage applies an exponential backoff strategy. When the maximum number of retries is hit, a :class:`.error.exceptions.RetryError` is raised
+        :param max_retries: maximum number of retries for the stage before raising a RetryError(SoftError) (default 0)
+        :param backoff: backoff factor for the `exponential backoff strategy` used by the stage when it raises one of the exceptions declared in `retryable_errors` param (default 0.0)
         """
         self._executors_ready = False
         # FIXME here we force a BatchStage to run on a thread, but we would leave it on the main thread
@@ -497,6 +547,7 @@ class Pipeline:
         if args is None:
             args = []
         self._check_stage_name(name)
+        self._check_retries_params(retryable_errors, max_retries, backoff)
         # if it is concurrent and it is the first stage, make the source working on a output queue
         if concurrency > 0 and not self._containers:
             self._enqueue_source = True
@@ -507,7 +558,15 @@ class Pipeline:
 
         def append_stage(stage_future: Future):
             stage = stage_future.result()
-            container = self._build_container(name, stage, concurrency, parallel)
+            container = self._build_container(
+                name,
+                stage,
+                concurrency,
+                parallel,
+                retryable_errors,
+                max_retries,
+                backoff,
+            )
             self._wait_for_previous(container, last_stage_name)
             self._containers[name] = container
 
@@ -557,6 +616,31 @@ class Pipeline:
         if name in self._containers:
             raise ValueError(f"The stage name {name} is already used in this pipeline")
 
+    @staticmethod
+    def _check_retries_params(
+        retryable_errors: Tuple[Type[Exception], ...],
+        max_retries: int,
+        backoff: Union[float, int],
+    ):
+        """
+        Check the values used to apply the retry strategy for a stage
+        """
+        if not (isinstance(backoff, (float, int)) and backoff >= 0):
+            raise ValueError(
+                "The `backoff` parameter must be either a float or int and its value must be >= 0"
+            )
+        if not (isinstance(max_retries, int) and max_retries >= 0):
+            raise ValueError(
+                "The `max_retries` parameter must be an int and its value must be >= 0"
+            )
+        if not (
+            isinstance(retryable_errors, tuple)
+            and all(issubclass(err, Exception) for err in retryable_errors)
+        ):
+            raise ValueError(
+                "The `retryable_errors` parameter must be a tuple and all its elements must be Exceptions"
+            )
+
     def _init_out_queue(self):
         """
         Get the internal output pipeline for asynchronous processing of single items
